#
# Generated by TigerShark.tools.convertPyX12 on 2012-07-10 16:29:59.666466
#
from tigershark.X12.parse import Message, Loop, Segment, Composite, Element, Properties
parsed_820_1000A = Loop( u'1000A', Properties(looptype='',repeat=u'1',pos=u'070',req_sit=u'R',desc=u"Premium Receiver's Name"),
Segment( u'N1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'070',desc=u"Premium Receiver's Name"),
  Element( u'N101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'PE'] ) ),
  Element( u'N102', Properties(desc=u'Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
  Element( u'N103', Properties(desc=u'Identification Code Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'1', u'9', u'EQ', u'FI', u'XV'] ) ),
  Element( u'N104', Properties(desc=u'Identification Code', req_sit=u'S', data_type=(u'AN',u'2',u'80'), position=4,
    codes=[] ) ),
  Element( u'N105', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[] ) ),
  Element( u'N106', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=6,
    codes=[] ) ),
),
Segment( u'N2', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'080',desc=u'Premium Receiver Additional Name'),
  Element( u'N201', Properties(desc=u'Name', req_sit=u'R', data_type=(u'AN',u'1',u'60'), position=1,
    codes=[] ) ),
  Element( u'N202', Properties(desc=u'Name', req_sit=u'N', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
),
Segment( u'N3', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'090',desc=u"Premium Receiver's Address"),
  Element( u'N301', Properties(desc=u'Address Information', req_sit=u'R', data_type=(u'AN',u'1',u'55'), position=1,
    codes=[] ) ),
  Element( u'N302', Properties(desc=u'Address Information', req_sit=u'S', data_type=(u'AN',u'1',u'55'), position=2,
    codes=[] ) ),
),
Segment( u'N4', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'100',desc=u"Premium Receiver's City, State, ZIP"),
  Element( u'N401', Properties(desc=u'City Name', req_sit=u'R', data_type=(u'AN',u'2',u'30'), position=1,
    codes=[] ) ),
  Element( u'N402', Properties(desc=u'State or Province Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[] ) ),
  Element( u'N403', Properties(desc=u'Postal Code', req_sit=u'R', data_type=(u'ID',u'3',u'15'), position=3,
    codes=[] ) ),
  Element( u'N404', Properties(desc=u'Country Code', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'N405', Properties(desc=u'Location Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=5,
    codes=[] ) ),
  Element( u'N406', Properties(desc=u'Location Identifier', req_sit=u'N', data_type=(u'AN',u'1',u'30'), position=6,
    codes=[] ) ),
),
)
parsed_820_1000B = Loop( u'1000B', Properties(looptype='',repeat=u'1',pos=u'070',req_sit=u'R',desc=u"Premium Payer's Name"),
Segment( u'N1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'070',desc=u"Premium Payer's Name"),
  Element( u'N101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'PR'] ) ),
  Element( u'N102', Properties(desc=u'Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
  Element( u'N103', Properties(desc=u'Identification Code Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'1', u'9', u'24', u'65', u'75', u'EQ', u'FI', u'PI'] ) ),
  Element( u'N104', Properties(desc=u'Identification Code', req_sit=u'S', data_type=(u'AN',u'2',u'80'), position=4,
    codes=[] ) ),
  Element( u'N105', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[] ) ),
  Element( u'N106', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=6,
    codes=[] ) ),
),
Segment( u'N2', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'080',desc=u'Premium Payer Additional Name'),
  Element( u'N201', Properties(desc=u'Name', req_sit=u'R', data_type=(u'AN',u'1',u'60'), position=1,
    codes=[] ) ),
  Element( u'N202', Properties(desc=u'Name', req_sit=u'N', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
),
Segment( u'N3', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'090',desc=u"Premium Payer's Address"),
  Element( u'N301', Properties(desc=u'Address Information', req_sit=u'R', data_type=(u'AN',u'1',u'55'), position=1,
    codes=[] ) ),
  Element( u'N302', Properties(desc=u'Address Information', req_sit=u'S', data_type=(u'AN',u'1',u'55'), position=2,
    codes=[] ) ),
),
Segment( u'N4', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'100',desc=u"Premium Payer's City, State, ZIP"),
  Element( u'N401', Properties(desc=u'City Name', req_sit=u'R', data_type=(u'AN',u'2',u'30'), position=1,
    codes=[] ) ),
  Element( u'N402', Properties(desc=u'State or Province Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[] ) ),
  Element( u'N403', Properties(desc=u'Postal Code', req_sit=u'R', data_type=(u'ID',u'3',u'15'), position=3,
    codes=[] ) ),
  Element( u'N404', Properties(desc=u'Country Code', req_sit=u'S', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'N405', Properties(desc=u'Location Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=5,
    codes=[] ) ),
  Element( u'N406', Properties(desc=u'Location Identifier', req_sit=u'N', data_type=(u'AN',u'1',u'30'), position=6,
    codes=[] ) ),
),
Segment( u'PER', Properties(syntax='',req_sit=u'S',repeat=u'>1',pos=u'120',desc=u"Premium Payer's Administrative Contact"),
  Element( u'PER01', Properties(desc=u'Contact Function Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'IC'] ) ),
  Element( u'PER02', Properties(desc=u'Name', req_sit=u'R', data_type=(u'AN',u'1',u'60'), position=2,
    codes=[] ) ),
  Element( u'PER03', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'EM', u'FX', u'TE'] ) ),
  Element( u'PER04', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=4,
    codes=[] ) ),
  Element( u'PER05', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[u'EM', u'EX', u'FX', u'TE'] ) ),
  Element( u'PER06', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=6,
    codes=[] ) ),
  Element( u'PER07', Properties(desc=u'Communication Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[u'EM', u'EX', u'FX', u'TE'] ) ),
  Element( u'PER08', Properties(desc=u'Communication Number', req_sit=u'S', data_type=(u'AN',u'1',u'256'), position=8,
    codes=[] ) ),
  Element( u'PER09', Properties(desc=u'Contact Inquiry Reference', req_sit=u'N', data_type=(u'AN',u'1',u'20'), position=9,
    codes=[] ) ),
),
)
parsed_820_HEADER = Loop( u'HEADER', Properties(looptype=u'wrapper',repeat=u'1',pos=u'015',req_sit=u'R',desc=u'Table 1 - Header'),
Segment( u'BPR', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'020',desc=u'Financial Information'),
  Element( u'BPR01', Properties(desc=u'Transaction Handling Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'C', u'D', u'I', u'P', u'U', u'X'] ) ),
  Element( u'BPR02', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=2,
    codes=[] ) ),
  Element( u'BPR03', Properties(desc=u'Credit/Debit Flag Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=3,
    codes=[u'C', u'D'] ) ),
  Element( u'BPR04', Properties(desc=u'Payment Method Code', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=4,
    codes=[u'ACH', u'BOP', u'CHK', u'FWT', u'SWT'] ) ),
  Element( u'BPR05', Properties(desc=u'Payment Format Code', req_sit=u'S', data_type=(u'ID',u'1',u'10'), position=5,
    codes=[u'CCP', u'CTX'] ) ),
  Element( u'BPR06', Properties(desc=u'(DFI) ID Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=6,
    codes=[u'01', u'04'] ) ),
  Element( u'BPR07', Properties(desc=u'(DFI) Identification Number', req_sit=u'S', data_type=(u'AN',u'3',u'12'), position=7,
    codes=[] ) ),
  Element( u'BPR08', Properties(desc=u'Account Number Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'3'), position=8,
    codes=[u'DA', u'ALC'] ) ),
  Element( u'BPR09', Properties(desc=u'Account Number', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=9,
    codes=[] ) ),
  Element( u'BPR10', Properties(desc=u'Originating Company Identifier', req_sit=u'S', data_type=(u'AN',u'10',u'10'), position=10,
    codes=[] ) ),
  Element( u'BPR11', Properties(desc=u'Originating Company Supplemental Code', req_sit=u'S', data_type=(u'AN',u'9',u'9'), position=11,
    codes=[] ) ),
  Element( u'BPR12', Properties(desc=u'(DFI) ID Number Qualifier', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=12,
    codes=[u'01', u'04'] ) ),
  Element( u'BPR13', Properties(desc=u'(DFI) Identification Number', req_sit=u'S', data_type=(u'AN',u'3',u'12'), position=13,
    codes=[] ) ),
  Element( u'BPR14', Properties(desc=u'Account Number Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'3'), position=14,
    codes=[u'DA', u'SG'] ) ),
  Element( u'BPR15', Properties(desc=u'Account Number', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=15,
    codes=[] ) ),
  Element( u'BPR16', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=16,
    codes=[] ) ),
  Element( u'BPR17', Properties(desc=u'Business Function Code', req_sit=u'N', data_type=(u'ID',u'1',u'3'), position=17,
    codes=[] ) ),
  Element( u'BPR18', Properties(desc=u'(DFI) ID Number Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=18,
    codes=[] ) ),
  Element( u'BPR19', Properties(desc=u'(DFI) Identification Number', req_sit=u'N', data_type=(u'AN',u'3',u'12'), position=19,
    codes=[] ) ),
  Element( u'BPR20', Properties(desc=u'Account Number Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'3'), position=20,
    codes=[] ) ),
  Element( u'BPR21', Properties(desc=u'Account Number', req_sit=u'N', data_type=(u'AN',u'1',u'35'), position=21,
    codes=[] ) ),
),
Segment( u'TRN', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'035',desc=u'Reassociation Key'),
  Element( u'TRN01', Properties(desc=u'Trace Type Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=1,
    codes=[u'1', u'3'] ) ),
  Element( u'TRN02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'TRN03', Properties(desc=u'Originating Company Identifier', req_sit=u'S', data_type=(u'AN',u'10',u'10'), position=3,
    codes=[] ) ),
  Element( u'TRN04', Properties(desc=u'Reference Identification', req_sit=u'S', data_type=(u'AN',u'1',u'50'), position=4,
    codes=[] ) ),
),
Segment( u'CUR', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'040',desc=u'Non-US Dollars Currency'),
  Element( u'CUR01', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'2B', u'PR'] ) ),
  Element( u'CUR02', Properties(desc=u'Currency Code', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=2,
    codes=[u'MXP', u'CAD', u'USD'] ) ),
  Element( u'CUR03', Properties(desc=u'Exchange Rate', req_sit=u'S', data_type=(u'R',u'4',u'10'), position=3,
    codes=[] ) ),
  Element( u'CUR04', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=4,
    codes=[] ) ),
  Element( u'CUR05', Properties(desc=u'Currency Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=5,
    codes=[] ) ),
  Element( u'CUR06', Properties(desc=u'Currency Market/Exchange Code', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=6,
    codes=[] ) ),
  Element( u'CUR07', Properties(desc=u'Date/Time Qualifier', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=7,
    codes=[] ) ),
  Element( u'CUR08', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=8,
    codes=[] ) ),
  Element( u'CUR09', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=9,
    codes=[] ) ),
  Element( u'CUR10', Properties(desc=u'Date/Time Qualifier', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=10,
    codes=[] ) ),
  Element( u'CUR11', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=11,
    codes=[] ) ),
  Element( u'CUR12', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=12,
    codes=[] ) ),
  Element( u'CUR13', Properties(desc=u'Date/Time Qualifier', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=13,
    codes=[] ) ),
  Element( u'CUR14', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=14,
    codes=[] ) ),
  Element( u'CUR15', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=15,
    codes=[] ) ),
  Element( u'CUR16', Properties(desc=u'Date/Time Qualifier', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=16,
    codes=[] ) ),
  Element( u'CUR17', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=17,
    codes=[] ) ),
  Element( u'CUR18', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=18,
    codes=[] ) ),
  Element( u'CUR19', Properties(desc=u'Date/Time Qualifier', req_sit=u'N', data_type=(u'ID',u'3',u'3'), position=19,
    codes=[] ) ),
  Element( u'CUR20', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=20,
    codes=[] ) ),
  Element( u'CUR21', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=21,
    codes=[] ) ),
),
Segment( u'REF', Properties(syntax='',req_sit=u'S',repeat=u'>1',pos=u'050',desc=u'Premium Receivers Identification Key'),
  Element( u'REF01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'14', u'18', u'2F', u'38', u'72'] ) ),
  Element( u'REF02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'REF03', Properties(desc=u'Description', req_sit=u'N', data_type=(u'AN',u'1',u'80'), position=3,
    codes=[] ) ),
  Composite( u'C040', Properties(req_sit=u'N',refdes='',seq=u'04',desc=u'Reference Identifier'),
  ),
),
Segment( u'DTM', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Process Date'),
  Element( u'DTM01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'009'] ) ),
  Element( u'DTM02', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=2,
    codes=[] ) ),
  Element( u'DTM03', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=3,
    codes=[] ) ),
  Element( u'DTM04', Properties(desc=u'Time Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Element( u'DTM05', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[] ) ),
  Element( u'DTM06', Properties(desc=u'Date Time Period', req_sit=u'N', data_type=(u'AN',u'1',u'35'), position=6,
    codes=[] ) ),
),
Segment( u'DTM', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Delivery Date'),
  Element( u'DTM01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'035'] ) ),
  Element( u'DTM02', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=2,
    codes=[] ) ),
  Element( u'DTM03', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=3,
    codes=[] ) ),
  Element( u'DTM04', Properties(desc=u'Time Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Element( u'DTM05', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[] ) ),
  Element( u'DTM06', Properties(desc=u'Date Time Period', req_sit=u'N', data_type=(u'AN',u'1',u'35'), position=6,
    codes=[] ) ),
),
Segment( u'DTM', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'060',desc=u'Coverage Period'),
  Element( u'DTM01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'582'] ) ),
  Element( u'DTM02', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=2,
    codes=[] ) ),
  Element( u'DTM03', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=3,
    codes=[] ) ),
  Element( u'DTM04', Properties(desc=u'Time Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Element( u'DTM05', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[u'RD8'] ) ),
  Element( u'DTM06', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=6,
    codes=[] ) ),
),
parsed_820_1000A,
parsed_820_1000B,
)
parsed_820_2315A = Loop( u'2315A', Properties(looptype='',repeat=u'>1',pos=u'204',req_sit=u'S',desc=u'Member Count'),
Segment( u'SLN', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'204',desc=u'Member Count'),
  Element( u'SLN01', Properties(desc=u'Assigned Identification', req_sit=u'R', data_type=(u'AN',u'1',u'20'), position=1,
    codes=[] ) ),
  Element( u'SLN02', Properties(desc=u'Assigned Identification', req_sit=u'N', data_type=(u'AN',u'1',u'20'), position=2,
    codes=[] ) ),
  Element( u'SLN03', Properties(desc=u'Relationship Code', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=3,
    codes=[u'O'] ) ),
  Element( u'SLN04', Properties(desc=u'Quantity', req_sit=u'R', data_type=(u'R',u'1',u'15'), position=4,
    codes=[] ) ),
  Composite( u'C001', Properties(req_sit=u'R',refdes='',seq=u'05',desc=u'Composite Unit of Measure'),
    Element( u'SLN05-01', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=0,
      codes=[u'10', u'IE', u'PR'] ) ),
    Element( u'SLN05-02', Properties(desc=u'Exponent', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=1,
      codes=[] ) ),
    Element( u'SLN05-03', Properties(desc=u'Multiplier', req_sit=u'N', data_type=(u'R',u'1',u'10'), position=2,
      codes=[] ) ),
    Element( u'SLN05-04', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=3,
      codes=[] ) ),
    Element( u'SLN05-05', Properties(desc=u'Exponent', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=4,
      codes=[] ) ),
    Element( u'SLN05-06', Properties(desc=u'Multiplier', req_sit=u'N', data_type=(u'R',u'1',u'10'), position=5,
      codes=[] ) ),
    Element( u'SLN05-07', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=6,
      codes=[] ) ),
    Element( u'SLN05-08', Properties(desc=u'Exponent', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=7,
      codes=[] ) ),
    Element( u'SLN05-09', Properties(desc=u'Multiplier', req_sit=u'N', data_type=(u'R',u'1',u'10'), position=8,
      codes=[] ) ),
    Element( u'SLN05-10', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=9,
      codes=[] ) ),
    Element( u'SLN05-11', Properties(desc=u'Exponent', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=10,
      codes=[] ) ),
    Element( u'SLN05-12', Properties(desc=u'Multiplier', req_sit=u'N', data_type=(u'R',u'1',u'10'), position=11,
      codes=[] ) ),
    Element( u'SLN05-13', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=12,
      codes=[] ) ),
    Element( u'SLN05-14', Properties(desc=u'Exponent', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=13,
      codes=[] ) ),
    Element( u'SLN05-15', Properties(desc=u'Multiplier', req_sit=u'N', data_type=(u'R',u'1',u'10'), position=14,
      codes=[] ) ),
  ),
  Element( u'SLN06', Properties(desc=u'Unit Price', req_sit=u'N', data_type=(u'R',u'1',u'17'), position=6,
    codes=[] ) ),
  Element( u'SLN07', Properties(desc=u'Basis of Unit Price Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[] ) ),
  Element( u'SLN08', Properties(desc=u'Relationship Code', req_sit=u'N', data_type=(u'ID',u'1',u'1'), position=8,
    codes=[] ) ),
  Element( u'SLN09', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=9,
    codes=[] ) ),
  Element( u'SLN10', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=10,
    codes=[] ) ),
  Element( u'SLN11', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=11,
    codes=[] ) ),
  Element( u'SLN12', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=12,
    codes=[] ) ),
  Element( u'SLN13', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=13,
    codes=[] ) ),
  Element( u'SLN14', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=14,
    codes=[] ) ),
  Element( u'SLN15', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=15,
    codes=[] ) ),
  Element( u'SLN16', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=16,
    codes=[] ) ),
  Element( u'SLN17', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=17,
    codes=[] ) ),
  Element( u'SLN18', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=18,
    codes=[] ) ),
  Element( u'SLN19', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=19,
    codes=[] ) ),
  Element( u'SLN20', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=20,
    codes=[] ) ),
  Element( u'SLN21', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=21,
    codes=[] ) ),
  Element( u'SLN22', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=22,
    codes=[] ) ),
  Element( u'SLN23', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=23,
    codes=[] ) ),
  Element( u'SLN24', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=24,
    codes=[] ) ),
  Element( u'SLN25', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=25,
    codes=[] ) ),
  Element( u'SLN26', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=26,
    codes=[] ) ),
  Element( u'SLN27', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=27,
    codes=[] ) ),
  Element( u'SLN28', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=28,
    codes=[] ) ),
),
)
parsed_820_2310A = Loop( u'2310A', Properties(looptype='',repeat=u'1',pos=u'190',req_sit=u'S',desc=u'Summary Line Item'),
Segment( u'IT1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'190',desc=u'Summary Line Item'),
  Element( u'IT101', Properties(desc=u'Assigned Identification', req_sit=u'R', data_type=(u'AN',u'1',u'20'), position=1,
    codes=[] ) ),
  Element( u'IT102', Properties(desc=u'Quantity Invoiced', req_sit=u'N', data_type=(u'R',u'1',u'15'), position=2,
    codes=[] ) ),
  Element( u'IT103', Properties(desc=u'Unit or Basis for Measurement Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[] ) ),
  Element( u'IT104', Properties(desc=u'Unit Price', req_sit=u'N', data_type=(u'R',u'1',u'17'), position=4,
    codes=[] ) ),
  Element( u'IT105', Properties(desc=u'Basis of Unit Price Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[] ) ),
  Element( u'IT106', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=6,
    codes=[] ) ),
  Element( u'IT107', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=7,
    codes=[] ) ),
  Element( u'IT108', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=8,
    codes=[] ) ),
  Element( u'IT109', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=9,
    codes=[] ) ),
  Element( u'IT110', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'IT111', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=11,
    codes=[] ) ),
  Element( u'IT112', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=12,
    codes=[] ) ),
  Element( u'IT113', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=13,
    codes=[] ) ),
  Element( u'IT114', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=14,
    codes=[] ) ),
  Element( u'IT115', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=15,
    codes=[] ) ),
  Element( u'IT116', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=16,
    codes=[] ) ),
  Element( u'IT117', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=17,
    codes=[] ) ),
  Element( u'IT118', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=18,
    codes=[] ) ),
  Element( u'IT119', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=19,
    codes=[] ) ),
  Element( u'IT120', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=20,
    codes=[] ) ),
  Element( u'IT121', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=21,
    codes=[] ) ),
  Element( u'IT122', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=22,
    codes=[] ) ),
  Element( u'IT123', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=23,
    codes=[] ) ),
  Element( u'IT124', Properties(desc=u'Product/Service ID Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=24,
    codes=[] ) ),
  Element( u'IT125', Properties(desc=u'Product/Service ID', req_sit=u'N', data_type=(u'AN',u'1',u'48'), position=25,
    codes=[] ) ),
),
parsed_820_2315A,
)
parsed_820_2320A = Loop( u'2320A', Properties(looptype='',repeat=u'>1',pos=u'210',req_sit=u'S',desc=u'Organization Summary Remittance Level Adjustment'),
Segment( u'ADX', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'210',desc=u'Organization Summary Remittance Level Adjustment'),
  Element( u'ADX01', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=1,
    codes=[] ) ),
  Element( u'ADX02', Properties(desc=u'Adjustment Reason Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[u'20', u'52', u'53', u'AA', u'H1', u'H6', u'IA', u'J3'] ) ),
  Element( u'ADX03', Properties(desc=u'Reference Identification Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=3,
    codes=[] ) ),
  Element( u'ADX04', Properties(desc=u'Reference Identification', req_sit=u'N', data_type=(u'AN',u'1',u'50'), position=4,
    codes=[] ) ),
),
)
parsed_820_2300A = Loop( u'2300A', Properties(looptype='',repeat=u'>1',pos=u'150',req_sit=u'R',desc=u'Organization Summary Remittance Detail'),
Segment( u'RMR', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'150',desc=u'Organization Summary Remittance Detail'),
  Element( u'RMR01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'11', u'1L', u'CT', u'IK'] ) ),
  Element( u'RMR02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'RMR03', Properties(desc=u'Payment Action Code', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'PA', u'PI', u'PO', u'PP'] ) ),
  Element( u'RMR04', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=4,
    codes=[] ) ),
  Element( u'RMR05', Properties(desc=u'Monetary Amount', req_sit=u'S', data_type=(u'R',u'1',u'18'), position=5,
    codes=[] ) ),
  Element( u'RMR06', Properties(desc=u'Monetary Amount', req_sit=u'N', data_type=(u'R',u'1',u'18'), position=6,
    codes=[] ) ),
  Element( u'RMR07', Properties(desc=u'Adjustment Reason Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[] ) ),
  Element( u'RMR08', Properties(desc=u'Monetary Amount', req_sit=u'N', data_type=(u'R',u'1',u'18'), position=8,
    codes=[] ) ),
),
parsed_820_2310A,
parsed_820_2320A,
)
parsed_820_2000A = Loop( u'2000A', Properties(looptype='',repeat=u'1',pos=u'010',req_sit=u'S',desc=u'Organization Summary Remittance'),
Segment( u'ENT', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Organization Summary Remittance'),
  Element( u'ENT01', Properties(desc=u'Assigned Number', req_sit=u'R', data_type=(u'N0',u'1',u'6'), position=1,
    codes=[] ) ),
  Element( u'ENT02', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'2L'] ) ),
  Element( u'ENT03', Properties(desc=u'Identification Code Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'1', u'9', u'FI'] ) ),
  Element( u'ENT04', Properties(desc=u'Identification Code', req_sit=u'S', data_type=(u'AN',u'2',u'80'), position=4,
    codes=[] ) ),
  Element( u'ENT05', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[] ) ),
  Element( u'ENT06', Properties(desc=u'Identification Code Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=6,
    codes=[] ) ),
  Element( u'ENT07', Properties(desc=u'Identification Code', req_sit=u'N', data_type=(u'AN',u'2',u'80'), position=7,
    codes=[] ) ),
  Element( u'ENT08', Properties(desc=u'Reference Identification Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=8,
    codes=[] ) ),
  Element( u'ENT09', Properties(desc=u'Reference Identification', req_sit=u'N', data_type=(u'AN',u'1',u'50'), position=9,
    codes=[] ) ),
),
parsed_820_2300A,
)
parsed_820_2100B = Loop( u'2100B', Properties(looptype='',repeat=u'>1',pos=u'020',req_sit=u'S',desc=u'Individual Name'),
Segment( u'NM1', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'020',desc=u'Individual Name'),
  Element( u'NM101', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'EY', u'QE'] ) ),
  Element( u'NM102', Properties(desc=u'Entity Type Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=2,
    codes=[u'1'] ) ),
  Element( u'NM103', Properties(desc=u'Name Last or Organization Name', req_sit=u'S', data_type=(u'AN',u'1',u'60'), position=3,
    codes=[] ) ),
  Element( u'NM104', Properties(desc=u'Name First', req_sit=u'S', data_type=(u'AN',u'1',u'35'), position=4,
    codes=[] ) ),
  Element( u'NM105', Properties(desc=u'Name Middle', req_sit=u'S', data_type=(u'AN',u'1',u'25'), position=5,
    codes=[] ) ),
  Element( u'NM106', Properties(desc=u'Name Prefix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=6,
    codes=[] ) ),
  Element( u'NM107', Properties(desc=u'Name Suffix', req_sit=u'S', data_type=(u'AN',u'1',u'10'), position=7,
    codes=[] ) ),
  Element( u'NM108', Properties(desc=u'Identification Code Qualifier', req_sit=u'S', data_type=(u'ID',u'1',u'2'), position=8,
    codes=[u'34', u'EI', u'N'] ) ),
  Element( u'NM109', Properties(desc=u'Identification Code', req_sit=u'S', data_type=(u'AN',u'2',u'80'), position=9,
    codes=[] ) ),
  Element( u'NM110', Properties(desc=u'Entity Relationship Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=10,
    codes=[] ) ),
  Element( u'NM111', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=11,
    codes=[] ) ),
),
)
parsed_820_2320B = Loop( u'2320B', Properties(looptype='',repeat=u'>1',pos=u'210',req_sit=u'S',desc=u'Individual Premium Adjustment'),
Segment( u'ADX', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'210',desc=u'Individual Premium Adjustment'),
  Element( u'ADX01', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=1,
    codes=[] ) ),
  Element( u'ADX02', Properties(desc=u'Adjustment Reason Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=2,
    codes=[u'20', u'52', u'53', u'AA', u'AX', u'H1', u'H6', u'IA', u'J3'] ) ),
  Element( u'ADX03', Properties(desc=u'Reference Identification Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=3,
    codes=[] ) ),
  Element( u'ADX04', Properties(desc=u'Reference Identification', req_sit=u'N', data_type=(u'AN',u'1',u'50'), position=4,
    codes=[] ) ),
),
)
parsed_820_2300B = Loop( u'2300B', Properties(looptype='',repeat=u'>1',pos=u'150',req_sit=u'S',desc=u'Individual Premium Remittance Detail'),
Segment( u'RMR', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'150',desc=u'Individual Premium Remittance Detail'),
  Element( u'RMR01', Properties(desc=u'Reference Identification Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=1,
    codes=[u'11', u'9J', u'AZ', u'B7', u'CT', u'ID', u'IG', u'IK', u'KW'] ) ),
  Element( u'RMR02', Properties(desc=u'Reference Identification', req_sit=u'R', data_type=(u'AN',u'1',u'50'), position=2,
    codes=[] ) ),
  Element( u'RMR03', Properties(desc=u'Payment Action Code', req_sit=u'S', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'PI', u'PP'] ) ),
  Element( u'RMR04', Properties(desc=u'Monetary Amount', req_sit=u'R', data_type=(u'R',u'1',u'18'), position=4,
    codes=[] ) ),
  Element( u'RMR05', Properties(desc=u'Monetary Amount', req_sit=u'S', data_type=(u'R',u'1',u'18'), position=5,
    codes=[] ) ),
  Element( u'RMR06', Properties(desc=u'Monetary Amount', req_sit=u'N', data_type=(u'R',u'1',u'18'), position=6,
    codes=[] ) ),
  Element( u'RMR07', Properties(desc=u'Adjustment Reason Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[] ) ),
  Element( u'RMR08', Properties(desc=u'Monetary Amount', req_sit=u'N', data_type=(u'R',u'1',u'18'), position=8,
    codes=[] ) ),
),
Segment( u'DTM', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'180',desc=u'Individual Coverage Period'),
  Element( u'DTM01', Properties(desc=u'Date/Time Qualifier', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'582'] ) ),
  Element( u'DTM02', Properties(desc=u'Date', req_sit=u'N', data_type=(u'DT',u'8',u'8'), position=2,
    codes=[] ) ),
  Element( u'DTM03', Properties(desc=u'Time', req_sit=u'N', data_type=(u'TM',u'4',u'8'), position=3,
    codes=[] ) ),
  Element( u'DTM04', Properties(desc=u'Time Code', req_sit=u'N', data_type=(u'ID',u'2',u'2'), position=4,
    codes=[] ) ),
  Element( u'DTM05', Properties(desc=u'Date Time Period Format Qualifier', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[u'RD8'] ) ),
  Element( u'DTM06', Properties(desc=u'Date Time Period', req_sit=u'R', data_type=(u'AN',u'1',u'35'), position=6,
    codes=[] ) ),
),
parsed_820_2320B,
)
parsed_820_2000B = Loop( u'2000B', Properties(looptype='',repeat=u'>1',pos=u'160',req_sit=u'S',desc=u'Individual Remittance'),
Segment( u'ENT', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Individual Remittance'),
  Element( u'ENT01', Properties(desc=u'Assigned Number', req_sit=u'R', data_type=(u'N0',u'1',u'6'), position=1,
    codes=[] ) ),
  Element( u'ENT02', Properties(desc=u'Entity Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'3'), position=2,
    codes=[u'2J'] ) ),
  Element( u'ENT03', Properties(desc=u'Identification Code Qualifier', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=3,
    codes=[u'34', u'EI', u'ZZ'] ) ),
  Element( u'ENT04', Properties(desc=u'Identification Code', req_sit=u'R', data_type=(u'AN',u'2',u'80'), position=4,
    codes=[] ) ),
  Element( u'ENT05', Properties(desc=u'Entity Identifier Code', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=5,
    codes=[] ) ),
  Element( u'ENT06', Properties(desc=u'Identification Code Qualifier', req_sit=u'N', data_type=(u'ID',u'1',u'2'), position=6,
    codes=[] ) ),
  Element( u'ENT07', Properties(desc=u'Identification Code', req_sit=u'N', data_type=(u'AN',u'2',u'80'), position=7,
    codes=[] ) ),
  Element( u'ENT08', Properties(desc=u'Reference Identification Qualifier', req_sit=u'N', data_type=(u'ID',u'2',u'3'), position=8,
    codes=[] ) ),
  Element( u'ENT09', Properties(desc=u'Reference Identification', req_sit=u'N', data_type=(u'AN',u'1',u'50'), position=9,
    codes=[] ) ),
),
parsed_820_2100B,
parsed_820_2300B,
)
parsed_820_DETAIL = Loop( u'DETAIL', Properties(looptype=u'wrapper',repeat=u'>1',pos=u'020',req_sit=u'S',desc=u'Table 2 - Detail'),
parsed_820_2000A,
parsed_820_2000B,
)
parsed_820_ST_LOOP = Loop( u'ST_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'020',req_sit=u'R',desc=u'Transaction Set Header'),
Segment( u'ST', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'820 Header'),
  Element( u'ST01', Properties(desc=u'Transaction Set Identifier Code', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=1,
    codes=[u'820'] ) ),
  Element( u'ST02', Properties(desc=u'Transaction Set Control Number', req_sit=u'R', data_type=(u'AN',u'4',u'9'), position=2,
    codes=[] ) ),
),
parsed_820_HEADER,
parsed_820_DETAIL,
Segment( u'SE', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'050',desc=u'820 Trailer'),
  Element( u'SE01', Properties(desc=u'Number of Included Segments', req_sit=u'R', data_type=(u'N0',u'1',u'10'), position=1,
    codes=[] ) ),
  Element( u'SE02', Properties(desc=u'Transaction Set Control Number', req_sit=u'R', data_type=(u'AN',u'4',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_820_GS_LOOP = Loop( u'GS_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'020',req_sit=u'R',desc=u'Functional Group Header'),
Segment( u'GS', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Functional Group Header'),
  Element( u'GS01', Properties(desc=u'Functional Identifier Code', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'RA'] ) ),
  Element( u'GS02', Properties(desc=u'Application Senders Code', req_sit=u'R', data_type=(u'AN',u'2',u'15'), position=2,
    codes=[] ) ),
  Element( u'GS03', Properties(desc=u'124', req_sit=u'R', data_type=(u'AN',u'2',u'15'), position=3,
    codes=[] ) ),
  Element( u'GS04', Properties(desc=u'Date', req_sit=u'R', data_type=(u'DT',u'8',u'8'), position=4,
    codes=[] ) ),
  Element( u'GS05', Properties(desc=u'Time', req_sit=u'R', data_type=(u'TM',u'4',u'8'), position=5,
    codes=[] ) ),
  Element( u'GS06', Properties(desc=u'Group Control Number', req_sit=u'R', data_type=(u'N0',u'1',u'9'), position=6,
    codes=[] ) ),
  Element( u'GS07', Properties(desc=u'Responsible Agency Code', req_sit=u'R', data_type=(u'ID',u'1',u'2'), position=7,
    codes=[u'X'] ) ),
  Element( u'GS08', Properties(desc=u'Version / Release / Industry Identifier Code', req_sit=u'R', data_type=(u'AN',u'1',u'12'), position=8,
    codes=[u'004010X061A1'] ) ),
),
parsed_820_ST_LOOP,
Segment( u'GE', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Functional Group Trailer'),
  Element( u'GE01', Properties(desc=u'97', req_sit=u'R', data_type=(u'N0',u'1',u'6'), position=1,
    codes=[] ) ),
  Element( u'GE02', Properties(desc=u'Group Control Number', req_sit=u'R', data_type=(u'N0',u'1',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_820_ISA_LOOP = Loop( u'ISA_LOOP', Properties(looptype=u'explicit',repeat=u'>1',pos=u'001',req_sit=u'R',desc=u'Interchange Control Header'),
Segment( u'ISA', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'010',desc=u'Interchange Control Header'),
  Element( u'ISA01', Properties(desc=u'I01', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=1,
    codes=[u'00', u'03'] ) ),
  Element( u'ISA02', Properties(desc=u'I02', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=2,
    codes=[] ) ),
  Element( u'ISA03', Properties(desc=u'I03', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=3,
    codes=[u'00', u'01'] ) ),
  Element( u'ISA04', Properties(desc=u'I04', req_sit=u'R', data_type=(u'AN',u'10',u'10'), position=4,
    codes=[] ) ),
  Element( u'ISA05', Properties(desc=u'I05', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=5,
    codes=[u'01', u'14', u'20', u'27', u'28', u'29', u'30', u'33', u'ZZ'] ) ),
  Element( u'ISA06', Properties(desc=u'I06', req_sit=u'R', data_type=(u'AN',u'15',u'15'), position=6,
    codes=[] ) ),
  Element( u'ISA07', Properties(desc=u'I05', req_sit=u'R', data_type=(u'ID',u'2',u'2'), position=7,
    codes=[u'01', u'14', u'20', u'27', u'28', u'29', u'30', u'33', u'ZZ'] ) ),
  Element( u'ISA08', Properties(desc=u'I07', req_sit=u'R', data_type=(u'AN',u'15',u'15'), position=8,
    codes=[] ) ),
  Element( u'ISA09', Properties(desc=u'I08', req_sit=u'R', data_type=(u'DT',u'6',u'6'), position=9,
    codes=[] ) ),
  Element( u'ISA10', Properties(desc=u'I09', req_sit=u'R', data_type=(u'TM',u'4',u'4'), position=10,
    codes=[] ) ),
  Element( u'ISA11', Properties(desc=u'I10', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=11,
    codes=[u'U'] ) ),
  Element( u'ISA12', Properties(desc=u'I11', req_sit=u'R', data_type=(u'ID',u'5',u'5'), position=12,
    codes=[u'00401'] ) ),
  Element( u'ISA13', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=13,
    codes=[] ) ),
  Element( u'ISA14', Properties(desc=u'I13', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=14,
    codes=[u'0', u'1'] ) ),
  Element( u'ISA15', Properties(desc=u'I14', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=15,
    codes=[u'P', u'T'] ) ),
  Element( u'ISA16', Properties(desc=u'I15', req_sit=u'R', data_type=(u'AN',u'1',u'1'), position=16,
    codes=[] ) ),
),
parsed_820_GS_LOOP,
Segment( u'TA1', Properties(syntax='',req_sit=u'S',repeat=u'1',pos=u'020',desc=u'Interchange Acknowledgement'),
  Element( u'TA101', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=1,
    codes=[] ) ),
  Element( u'TA102', Properties(desc=u'I08', req_sit=u'R', data_type=(u'DT',u'6',u'6'), position=2,
    codes=[] ) ),
  Element( u'TA103', Properties(desc=u'I09', req_sit=u'R', data_type=(u'TM',u'4',u'4'), position=3,
    codes=[] ) ),
  Element( u'TA104', Properties(desc=u'I17', req_sit=u'R', data_type=(u'ID',u'1',u'1'), position=4,
    codes=[u'A', u'E', u'R'] ) ),
  Element( u'TA105', Properties(desc=u'I18', req_sit=u'R', data_type=(u'ID',u'3',u'3'), position=5,
    codes=[u'000', u'001', u'002', u'003', u'004', u'005', u'006', u'007', u'008', u'009', u'010', u'011', u'012', u'013', u'014', u'015', u'016', u'017', u'018', u'019', u'020', u'021', u'022', u'023', u'024', u'025', u'026', u'027', u'028', u'029', u'030', u'031'] ) ),
),
Segment( u'IEA', Properties(syntax='',req_sit=u'R',repeat=u'1',pos=u'030',desc=u'Interchange Control Trailer'),
  Element( u'IEA01', Properties(desc=u'I16', req_sit=u'R', data_type=(u'N0',u'1',u'5'), position=1,
    codes=[] ) ),
  Element( u'IEA02', Properties(desc=u'I12', req_sit=u'R', data_type=(u'N0',u'9',u'9'), position=2,
    codes=[] ) ),
),
)
parsed_820 = Message( u'820', Properties(desc=u'HIPAA Payment Deducted and Other Group Premium Payment for Insurance Products X061A1-820'),
parsed_820_ISA_LOOP,
)




#!/usr/bin/env python
"""X12.message contains classes that are the default definitions of parts of
an X12 message instance.

An X12 message instance is a recursive construction of Loops, sub-Loops and Segments.
The leaf elements are always Segments.  For transmission, the Segment
leaves are flattened out; the loop structure is lost.

An X12 message instance is built by :mod:`X12.parse`.

Navigation
==========

The X12 message structure is a deeply nested collection of Loops and Segments.
Consquently, XPath-like functionality is helpful.  Literal XPath is a pain to
parse and process. However, a Pythonic set of methods and keyword arguments
can provide XPath-like features.

Each class, therefore has the following methods that correspond to XPath
axes: parent, child, ancestor, descendant.  [Optional: methods like following, preceding,
following_sibling, preceding_sibling, self, ancestor_or_self, descendent_or_self.]

Each method accepts a "node test" as a positional parameter.  Values are
simply the node type name, either Loop or Segment.

Each method has optional keywords which contains basically one expression.
the name="X" keyword is like :samp:`[@name=X]` in XPath.  The pos=n keyword is like :samp:`[n]`
in XPath.

Implementations
===============

Note that there are several alternative implementations:

-   This module, :mod:`X12.message`, which is pure Python.

-   The :mod:`web.claims.models` implementation, which is built using the
    Django ORM.

-   The :mod:`web.claims_837.models` implementation which has a
    complete set of X12Segments
    that reflect the various actual segment types used in :samp:`837`.

An :mod:`X12.message.sa` module could provde SQLAlchemy
mappings for the pure Python class definitions.

Consequently, a :class:`Factory` object is essential for building the correct
objects each alternative implementation.

Module Content
==============

This module contains the following class definitions.

    - :class:`Factory`.  Factory which builds object instances.

    - :class:`X12Structure`. Superclass for the resulting X12 message structure.

        - :class:`X12Message`.  The overall message -- a collection of Loops.

        - :class:`X12Loop`. A Loop is a collection of Segments and sub-Loops.

        - :class:`X12Segment`. A Segment is a collection of Elements and Composites.
          Segments names are essentially a type name; segments can be reused
          in a single Loop with different semantics.

        - X12Composite.  A Composite is a collection of Elements.  This is
          not actually implemented.

        - X12Element.  An atomic piece of data.  These are identified
          positionally within a segment's source text.

The generic definition of :class:`X12Segment` depends on having the :mod:`X12.parse`
definitions of the Segment,
Composite and Element available to interpret the strings of data associated
with each X12Segment.

If we want a more sophisticated, admin-friendly implementation, we have
to subclass :class:`X12Segment` for each unique kind of segment found in the X12
message definitions.  This is done in the Django and SQLAlchemy versions.

Since a Loop is just a kind of PK and sequence number
used to marshall a Message (and interpret the meaning of Segments), we
don't need to subclass Loop.

It's simplest to persist at the Segment level, with enough FK's and Occurance
identifiers to reconstruct a Loop from a sequence of Segments.

Details
=============

..  autoclass:: Factory
    :members:

..  autoclass:: X12Structure
    :members:

..  autoclass:: X12Loop
    :members:

..  autoclass:: X12Segment
    :members:

..  autoclass:: X12Message
    :members:

SQL Alchemy module
===========================

..  automodule:: X12.message.sa
"""

class Factory( object ):
    """Create the generic X12Message structures.

    This object is used by an :mod:`X12.parse` to unmarshall message
    text and create generic Python objects.

    Subclasses include

        - :class:`X12.message.sa` creates SQLAlchemy objects; these have endless
          subclasses for X12Segment.

        - :mod:`web.claims.models` creates Django objects.  The generic
          claims application uses a generic X12Segment implementation.

        - XXX future: :mod:`web.claims_837.models` will create Django objects;
          this version will have endless peer classes for X12Segment that
          carry the detailed element definitions.

    The "subclasses" are entirely a theoretical construct.  We don't *need*
    a properly polymorphic class hierarchy.
    """
    @staticmethod
    def makeMessage( name, *structure ):
        """Create a new X12Message object.
        :param name: name of this Message
        :param structure: the various Loops of this Message
        :returns: X12Message instance
        """
        return X12Message( name, *structure )
    @staticmethod
    def makeLoop( name, *structure ):
        """Create a new X12Loop object.
        :param name: name of this Loop
        :param structure: the various sub-Loops and Segments of this Loop
        :returns: X12Loop instance
        """
        return X12Loop( name, *structure )
    @staticmethod
    def makeSegment( segmentToken, compositeSep=":", segmentType=None ):
        """Create a new X12Segment object.
        :param segmentToken: the :class:`X12.parse.SegmentToken` created by the parser.
        :param compositeSep: Composite internal separator from the ISA segment.
        :param segmentType: an instance of :class:`X12.parse.Segment` which defines
        the Elements and Composites of this Segment.
        :returns: X12Segment instance
        """
        return X12Segment( segmentToken, segmentType )

class X12Structure( object ):
    """A piece of an X12 message structure.  This is the common superclass
    for X12Loop and X12Segment, allowing a recursive structure.

    :ivar name: name of this structure.
    :ivar thePosition: relative position within the parent.
    :ivar theParent: parent structure.
    :ivar occurrence: occurrence number for Loops or Segments which repeat.
    """
    def __init__( self, name ):
        self.name= name
        self.thePosition= 0
        self.theParent= None
        self.occurrence= 0
        self.theMessage= None
    def segs( self ):
        """Returns a flat list of Segments under this structure."""
        return NotImplemented
    def getMessage( self ):
        return self.theMessage
    def setMessage( self, aMessage ):
        self.theMessage= aMessage
    message= property( getMessage, setMessage )
    def matches( self, test, name=None, pos= None ):
        """The Xpath :samp:`self` axis"""
        testType= _test2type[test.lower()]
        if isinstance(self, testType ):
            c1, c2 = True, True
            if name is not None:
                c1= self.name == name
            if pos is not None:
                c2= self.thePosition == pos
            return c1 and c2
        return False
    def child( self, test, name=None, pos=None ):
        return NotImplemented
    def descendant( self, test, name=None, pos=None ):
        return NotImplemented
    def parent( self, test, name=None, pos=None ):
        """Examine the immediate parent of this node."""
        result= []
        if self.theParent.matches( test, name, pos ):
          result.append( self.theParent )
        return result
    def ancestor( self, test, name=None, pos=None ):
        """Examine nay ancestor of this node."""
        result= []
        if self.theParent:
            if self.theParent.matches( test, name, pos ):
                result.append( self.theParent )
            result.extend( self.theParent.ancestor(test,name,pos) )
        return result

class X12Loop( X12Structure ):
    """A Loop in an X12 message; it contains subloops and segments.
    This also the superclass for a Message, which includes Loops
    (but no Segments.)
    :ivar parts: sequence of component parts (subLoops or Segments)
    """
    def __init__( self, name, *parts ):
        """Create an X12Loop with a given name and list of X12Structure parts.
        :param name: name of this loop
        :param parts: (optional) list of parts to attach to this loop.
        """
        super(X12Loop,self).__init__( name )
        self.parts= []
        for p in parts:
            self.addChild( p )
    def __str__( self ):
        return "%s(%s)" % (
            self.name,
            ", ".join(map(str,self.parts)) )
    def getMessage( self ):
        return self.theMessage
    def setMessage( self, aMessage ):
        self.theMessage= aMessage
        for p in self.parts:
            p.message= aMessage
    message= property( getMessage, setMessage )
    def segs( self ):
        """Create a flat list of Segments under this Loop."""
        segList= []
        for p in self.parts:
            segList.extend( p.segs() )
        return segList
    def addChild( self, aPart ):
        """Add a child Loop or Segment to this Loop.
        Under some circumstances, we may need to propagate the message
        definition down into the lower levels of this structure.

        Note that this sets the message of the parts to the message
        associated with this Loop.  The message setting is recursive
        and assures that all subloops are also correctly associated with
        this message.
        """
        aPart.message= self.message
        aPart.theParent= self
        aPart.thePosition= len(self.parts) + 1
        self.parts.append( aPart )
    def child( self, test, name=None, pos=None ):
        """Examine immediate children of this node."""
        result= []
        for part in self.parts:
            if part.matches( test, name, pos ):
                result.append( part )
        return result
    def descendant( self, test, name=None, pos=None ):
        """Examine descendents of this node."""
        result= []
        for part in self.parts:
            if part.matches( test, name, pos ):
                result.append( part )
            result.extend( part.descendant(test,name,pos) )
        return result

class X12Segment( X12Structure ):
    """A segment in an X12 message, created by a parser from a segmentToken.
    :ivar elements: an instance of :class:`X12.parse.SegmentToken` which contains
    the individual Element values.
    :ivar segType: An optional instance of :class:`X12.parse.Segment` that provides
    names, data type properties for each Element of the Segment.
    """
    def __init__( self, segmentToken, segmentType=None ):
        """Create a Segment from a SegmentToken and a X12.parse.Segment definition.
        :param segmentToken: An :class:`X12.parse.SegmentToken` instance:
        a list-like collection of Element values.  It turns out that a simple
        list of values may also work, if it does NOT have trailing empty
        items omitted.  Real Segment Tokens can have trailing empty items
        omitted.
        :param segmentType: An :class:`X12.parse.Segment` instance, which
        defines the Elements and Composites of this X12Segment.
        """
        super(X12Segment,self).__init__( segmentToken[0] )
        self.elements= segmentToken
        self.segType= segmentType
    def __str__( self ):
        return str( self.elements )
    def bind( self, segment ):
        """Bind an :class:`X12.parse.Segment` to this data.
        :param segment: An :class:`X12.parse.Segment` instance
        """
        self.segType= segment
    def segs( self ):
        """Return this segment as a flat list of segments."""
        return [ self ]
    def getByPos( self, position ):
        """Get an element from this segment, by position.
        Positions appear to be numbered from 1.  Position 0 is
        the segment's name.
        :param position: the position of the element.
        :returns: string value in this position.  No type conversion is done.
        """
        return self.elements[int(position)]
    def setByPos( self, position, value ):
        """Set an element of this segment, by position.
        Positions appear to be numbered from 1.  Position 0 is
        the segment's name.
        :param position: the position of the element.
        :param value: String value to assign.  No type conversion is done.
        """
        while len(self.elements) <= int(position):
            self.elements.append("")
        self.elements[int(position)]= str(value)
    def positionOf( self, name ):
        """Locate the position of an element by name.
        :param name: name of the element.
        :returns: numberic position.
        """
        if self.segType is None:
            raise NotImplementedError( "No Segment bound to this data." )
        for en in range(len(self.segType.structure)):
            if name == self.segType.structure[en].name:
                return en+1
        raise ValueError( "Name %s Is Unknown in Segment %s" % ( name, self.name ))
    def getByName( self, name ):
        """Get an element from this segment, by name.
        :param name: the name of the element.
        :returns: string value in the element's position.  No type conversion is done.
        """
        return self.getByPos(self.positionOf(name))
    def child( self, test, name=None, pos=None ):
        """Examine immediate children of this node;
        segments have no children; the result is an empty list."""
        return []
    def descendant( self, test, name=None, pos=None ):
        """Examine all descendants of this node;
        segments have no children; the result is an empty list."""
        return []

class X12Message( X12Loop ):
    """An overall X12 Message: essentially a collection of Loops."""
    def __init__( self, *args ):
        super( X12Message, self ).__init__( *args )
        self.message= self
    def marshall( self, segSep="~", eltSep="*" ):
        """Marshall this message into an X12 string.
        :param segSep: (optional) Segment Terminator, usually :samp:`~`
        :param eltSep: (optional) Element Separator, usually :samp:`*`
        :returns: String with this X12 message marshalled
        """
        segments= self.segs()
        return segSep.join( [ eltSep.join( s.elements ) for s in segments ] ) + segSep
    def getCompositeSeparator( self ):
        """Returns the composite separator in :samp:`ISA16`.
        :returns: string value of the composite separator, often a :samp:`:`.
        """
        isaLoop= self.descendant( "segment", "ISA" )[0]
        return isaLoop.getByPos(16)

_test2type = { "loop" : X12Loop, "segment" : X12Segment, "message" : X12Message }



from decimal import Decimal

from tigershark.facade import X12LoopBridge
from tigershark.facade import X12SegmentBridge
from tigershark.facade import ElementAccess
from tigershark.facade import ElementSequenceAccess
from tigershark.facade import CompositeAccess
from tigershark.facade import SegmentSequenceAccess
from tigershark.facade import SegmentConversion
from tigershark.facade import D8
from tigershark.facade import Money
from tigershark.facade import Facade
from tigershark.facade import enum
from tigershark.facade.common import ClaimAdjustment
from tigershark.facade.common import ContactDetails
from tigershark.facade.common import NamedEntity
from tigershark.facade.enums.remittance import remittance_advice_codes
from tigershark.facade.utils import first


class Header(X12LoopBridge):
    loopName = "HEADER"

    class _FinancialInformation(X12LoopBridge):
        """ Use this to address a single payment to a single payee.

        A payee may represent a single provider, a provider group, or multiple
        providers in a chain. The BPR contains mandatory information, even
        when it is not being used to move funds electronically.
        """
        transaction_type = ElementAccess("BPR", 1, x12type=enum({
            "C": "Payment Accompanies Remittance Advice",
            "D": "Make Payment Only",
            "H": "Notification Only",
            "I": "Remittance Information Only",
            "P": "Prenotification of Ruture Transfers",
            "U": "Split Payment and Remittance",
            "X": "Handling Party's Option to Split Payment and Remittance."}))
        amount = ElementAccess("BPR", 2, x12type=Money)
        credit_or_debit = ElementAccess("BPR", 3, x12type=enum({
            "C": "Credit",
            "D": "Debit"}))
        payment_method = ElementAccess("BPR", 4, x12type=enum({
            "ACH": "Automated Clearing House (ACH)",
            "BOP": "Financial Institution Option",
            "CHK": "Check",
            "FWT": "Federal Reserve Funds/Wire Transfer - Nonrepetitive",
            "NON": "Non-payment Data (only for transaction-type=H)"}))
        payment_format = ElementAccess("BPR", 5, x12type=enum({
            "CCP": "Cash Concentration/Disbursement Plus Addenda",
            "CTX": "Coporate Trade Exchange"}))
        sender_aba_transit_routing_number = ElementAccess(
            "BPR", 7, qualifier=(6, "01"))
        sender_canadian_bank_branch_and_institution_number = ElementAccess(
            "BPR", 7, qualifier=(6, "04"))
        sender_account_type = ElementAccess("BPR", 8, x12type=enum({
            "DA": "Demand Deposit"}))
        sender_bank_account_number = ElementAccess("BPR", 9)
        sender_id = ElementAccess("BPR", 10)
        sender_supplemental_id = ElementAccess("BPR", 11)

        receiver_aba_transit_routing_number = ElementAccess(
            "BPR", 13, qualifier=(12, "01"))
        receiver_canadian_bank_branch_and_institution_number = ElementAccess(
            "BPR", 13, qualifier=(12, "04"))
        receiver_account_type = ElementAccess("BPR", 14, x12type=enum({
            "DA": "Demand Deposit",
            "SG": "Savings"}))
        receiver_bank_account_number = ElementAccess("BPR", 15)

        issue_date = ElementAccess("BPR", 16, x12type=D8)

    class _ReassociationTraceNumber(X12LoopBridge):
        """ Uniquely identify this transaction set.

        Also aid in reassociating payments and remittances that have been
        separated.
        """
        trace_type = ElementAccess("TRN", 1, x12type=enum({
            "1": "Current Transaction Trace Numbers"}))
        check_or_eft_trace_number = ElementAccess("TRN", 2)
        payer_id = ElementAccess("TRN", 3)
        originating_company_supplemental_code = ElementAccess("TRN", 4)

    class _CurrencyInformation(X12LoopBridge):
        """ Specify the currency and exchange rate, if applicable.

        Only used when the payment is not being made in United States Dollars
        or when the payment is made in a currency different from that in the
        original claim."""
        currency_code = ElementAccess("CUR", 2)
        exchange_rate = ElementAccess("CUR", 3)

    class _Receiver(X12LoopBridge):
        """ Used when the receiver of the transaction is not the payee. """
        id = ElementAccess("REF", 2, qualifier=(1, "EV"))
        description = ElementAccess("REF", 3, qualifier=(1, "EV"))
        reference_id = ElementAccess("REF", 4, qualifier=(1, "EV"))

    # TODO tests
    class _Version(X12LoopBridge):
        """ Reports the version number of the adjudication system which
        generated the claim payments.

        This is *NOT* the version number reported in the GS segment of the
        underlying EDI file. This is sometimes used for customer service
        requests, to help track down where an error was introduced."""
        version_number = ElementAccess("REF", 2, qualifier=(1, "F2"))
        description = ElementAccess("REF", 3, qualifier=(1, "F2"))
        reference_id = ElementAccess("REF", 4, qualifier=(1, "F2"))

    class _ProductionDate(X12LoopBridge):
        """ Production date of the claim.

        This *MUST* be supplied when the cutoff date of the adjudication
        system is different from the date of the 835."""
        date = ElementAccess("DTM", 2, qualifier=(1, "405"), x12type=D8)

    def __init__(self, aLoop, *args, **kwargs):
        super(Header, self).__init__(aLoop, *args, **kwargs)
        self.financial_information = Header._FinancialInformation(
            aLoop, *args, **kwargs)
        self.reassociation_trace_number = Header._ReassociationTraceNumber(
            aLoop, *args, **kwargs)
        self.currency_information = Header._CurrencyInformation(
            aLoop, *args, **kwargs)
        self.receiver = Header._Receiver(aLoop, *args, **kwargs)
        self.version = Header._Version(aLoop, *args, **kwargs)
        self.production_date = Header._ProductionDate(aLoop, *args, **kwargs)


class ProviderAdjustments(X12SegmentBridge):
    provider_id = ElementAccess("PLB", 1)
    date = ElementAccess("PLB", 2, x12type=D8)
    reason_1 = ElementAccess("PLB", 3)
    amount_1 = ElementAccess("PLB", 4, x12type=Money)
    reason_2 = ElementAccess("PLB", 5)
    amount_2 = ElementAccess("PLB", 6, x12type=Money)
    reason_3 = ElementAccess("PLB", 7)
    amount_3 = ElementAccess("PLB", 8, x12type=Money)
    reason_4 = ElementAccess("PLB", 9)
    amount_4 = ElementAccess("PLB", 10, x12type=Money)
    reason_5 = ElementAccess("PLB", 11)
    amount_5 = ElementAccess("PLB", 12, x12type=Money)
    reason_6 = ElementAccess("PLB", 13)
    amount_6 = ElementAccess("PLB", 14, x12type=Money)


class Footer(Facade, X12LoopBridge):
    loopName = "FOOTER"

    provider_adjustments = SegmentSequenceAccess(
        "PLB", x12type=SegmentConversion(ProviderAdjustments))


class Payer(X12LoopBridge):
    """Payer information from the 1000A loop."""
    loopName = "1000A"
    payer_id = ElementAccess("REF", 2, qualifier=(1, "2U"))
    submitter_id = ElementAccess("REF", 2, qualifier=(1, "EO"))
    # health industry number
    hin = ElementAccess("REF", 2, qualifier=(1, "HI"))
    naic_code = ElementAccess("REF", 2, qualifier=(1, "NF"))

    def __str__(self):
        return "%s, %s %s %s" % (self.last, self.first, self.mid, self.suffix)

    def __init__(self, aLoop, *args, **kwargs):
        super(Payer, self).__init__(aLoop, *args, **kwargs)
        self.contact_details = ContactDetails(aLoop, *args, **kwargs)


class Payee(X12LoopBridge):
    """Payee information from 1000B loop."""
    loopName = "1000B"
    state_license = ElementAccess("REF", 2, qualifier=(1, "0B"))
    provider_UPIN = ElementAccess("REF", 2, qualifier=(1, "1G"))
    pharmacy_number = ElementAccess("REF", 2, qualifier=(1, "D3"))
    payee_id = ElementAccess("REF", 2, qualifier=(1, "PQ"))
    tax_id = ElementAccess("REF", 2, qualifier=(1, "TJ"))

    def __init__(self, aLoop, *args, **kwargs):
        super(Payee, self).__init__(aLoop, *args, **kwargs)
        self.contact_details = ContactDetails(aLoop, *args, **kwargs)


class ClaimsOverview(X12LoopBridge):
    loopName = "2000"
    number = ElementAccess("LX", 1)
    provider_id = ElementAccess("TS3", 1)
    facility_type_code = ElementAccess("TS3", 2)
    fiscal_period_end = ElementAccess("TS3", 3, x12type=D8)
    claim_count = ElementAccess("TS3", 4)
    total_claim_charge = ElementAccess("TS3", 5, x12type=Money)
    # All of the below are optional and a 0/None value does not
    # neccessarily mean that value is *actually* 0
    total_covered_charge = ElementAccess("TS3", 6, x12type=Money)
    total_noncovered_charge = ElementAccess("TS3", 7, x12type=Money)
    total_denied_charge = ElementAccess("TS3", 8, x12type=Money)
    total_provider_payment = ElementAccess("TS3", 9, x12type=Money)
    total_interest = ElementAccess("TS3", 10, x12type=Money)
    total_contractual_adjustment = ElementAccess("TS3", 11, x12type=Money)
    total_gramm_rudman_reduction = ElementAccess("TS3", 12, x12type=Money)
    total_msp_payer = ElementAccess("TS3", 13, x12type=Money)
    total_blood_deductible = ElementAccess("TS3", 14, x12type=Money)
    total_non_lab_charge = ElementAccess("TS3", 15, x12type=Money)
    total_coinsurance = ElementAccess("TS3", 16, x12type=Money)
    total_hcpcs_reported_charge = ElementAccess("TS3", 17, x12type=Money)
    total_hspcs_payable = ElementAccess("TS3", 18, x12type=Money)
    total_deductible = ElementAccess("TS3", 19, x12type=Money)
    total_professional_component = ElementAccess("TS3", 20, x12type=Money)
    total_msp_patient_liability_met = ElementAccess("TS3", 21, x12type=Money)
    total_patient_reimbursement = ElementAccess("TS3", 22, x12type=Money)
    total_pip_claim = ElementAccess("TS3", 23, x12type=Money)
    total_pip_adjustment = ElementAccess("TS3", 24, x12type=Money)
    # TODO: TS2 segment


class Claim(Facade, X12LoopBridge):
    class _ServiceInfo(X12LoopBridge):
        """ Oh jeez, I'm so sorry about this mess.

        Each claim has several nested 2110 loops that have additional
        information pertaining to a claim. This is usually when the
        insurance company doesn't want to pay the whole amount billed.
        The only way I could think of to implement this is to do this
        mess of multiple inheritance.

        Rest assured I did not come to this decision lightly."""
        loopName = "2110"
        hcpcs_code = CompositeAccess("SVC", "HC", 1)
        charge = ElementAccess("SVC", 2, x12type=Money)
        payment = ElementAccess("SVC", 3, x12type=Money)
        quantity = ElementAccess("SVC", 5)
        start_date = ElementAccess("DTM", 2, qualifier=(1, "150"), x12type=D8)
        end_date = ElementAccess("DTM", 2, qualifier=(1, "151"), x12type=D8)
        service_date = ElementAccess(
            "DTM", 2, qualifier=(1, "472"), x12type=D8)

        # Identification
        apg_number = ElementAccess("REF", 2, qualifier=(1, "1S"))
        provider_control_number = ElementAccess("REF", 2, qualifier=(1, "6R"))
        authorization_number = ElementAccess("REF", 2, qualifier=(1, "BB"))
        attachment_code = ElementAccess("REF", 2, qualifier=(1, "E9"))
        prior_authorization_number = ElementAccess(
            "REF", 2, qualifier=(1, "G1"))
        predetermination_of_benefits_id = ElementAccess(
            "REF", 2, qualifier=(1, "G3"))
        location_number = ElementAccess("REF", 2, qualifier=(1, "LU"))
        rate_code_number = ElementAccess("REF", 2, qualifier=(1, "RB"))
        # TODO: Rendering provider info?

        # Line-item claim amounts
        allowed_amount = ElementAccess(
            "AMT", 2, qualifier=(1, "B6"), x12type=Money)
        per_day_limit = ElementAccess(
            "AMT", 2, qualifier=(1, "DY"), x12type=Money)
        deduction_amount = ElementAccess(
            "AMT", 2, qualifier=(1, "KH"), x12type=Money)
        tax_amount = ElementAccess(
            "AMT", 2, qualifier=(1, "T"), x12type=Money)
        total_claim_before_taxes = ElementAccess(
            "AMT", 2, qualifier=(1, "T2"), x12type=Money)

        not_covered_quantity = ElementAccess("QTY", 2, qualifier=(1, "NE"))
        notes = ElementSequenceAccess(
            "LQ", 2, qualPos=1, inList=("HE",),
            x12type=enum(remittance_advice_codes, raw_unknowns=True))

        def __init__(self, anX12Message, *args, **kwargs):
            super(Claim._ServiceInfo, self).__init__(
                anX12Message, *args, **kwargs)
            self.claim_adjustments = Claim._ClaimAdjustments(anX12Message)

        def get_actual_allowed_amount(self):
            if self.allowed_amount == Decimal('0.00'):
                # Actually means that the allowed amount is the full billed amt
                return self.charge
            return self.allowed_amount

        def get_actual_deductible(self):
            return (self.claim_adjustments
                    .patient_responsibility.total_amount('1'))

        def get_actual_coinsurance(self):
            return (self.claim_adjustments
                    .patient_responsibility.total_amount('2'))

        def get_actual_copayment(self):
            return (self.claim_adjustments
                    .patient_responsibility.total_amount('3'))

    class _ClaimPaymentInfo(X12LoopBridge):
        patient_control_number = ElementAccess("CLP", 1)
        status_code = ElementAccess("CLP", 2, x12type=enum({
            "1": "Processed as Primary",
            "2": "Processed as Secondary",
            "3": "Processed as Tertiary",
            "4": "Denied",
            "19": "Processed as Primary, Forwarded to Additional Payer(s)",
            "20": "Processed as Secondary, Forwarded to Additional Payer(s)",
            "21": "Processed as Tertiary, Forwarded to Additional Payer(s)",
            "22": "Reversal of Previous Payment",
            "23": "Not Our Claim, Forwarded to Additional Payer(s)",
            "25": "Predetermination Pricing Only - No Payment"}))
        total_charge = ElementAccess("CLP", 3, x12type=Money)
        payment = ElementAccess("CLP", 4, x12type=Money)
        patient_responsibility = ElementAccess("CLP", 5, x12type=Money)
        claim_type = ElementAccess("CLP", 6, x12type=enum({
            "12": "Preferred Provider Organization (PPO)",
            "13": "Point of Service (POS)",
            "14": "Exclusive Provider Organization (EPO)",
            "15": "Indemnity Insurance",
            "16": "Health Maintenance Organization (HMO) Medicare Risk",
            "AM": "Automobile Medical",
            "CH": "Champus",
            "DS": "Disability",
            "HM": "Health Maintenance Organization",
            "LM": "Liability Medical",
            "MA": "Medicare Part A",
            "MB": "Medicare Part B",
            "MC": "Medicaid",
            "OF": "Other Federal Program",
            "TV": "Title V",
            "VA": "Veteran Administration Plan",
            "WC": "Workers Compensation Health Claim"}))
        payer_claim_control_number = ElementAccess("CLP", 7)
        facility_type = ElementAccess("CLP", 8)
        frequency_code = ElementAccess("CLP", 9)
        diagnosis_related_group_weight = ElementAccess("CLP", 11)
        discharge_fraction = ElementAccess("CLP", 12)
        # TODO: Medicare inpatient/outpatient adjudication?

        total_covered_charge = ElementAccess(
            "AMT", 2, qualifier=(1, "AU"), x12type=Money)
        discount_amount = ElementAccess(
            "AMT", 2, qualifier=(1, "D8"), x12type=Money)
        per_day_limit = ElementAccess(
            "AMT", 2, qualifier=(1, "DY"), x12type=Money)
        patient_amount_paid = ElementAccess(
            "AMT", 2, qualifier=(1, "F5"), x12type=Money)
        interest = ElementAccess(
            "AMT", 2, qualifier=(1, "I"), x12type=Money)
        negative_ledger = ElementAccess(
            "AMT", 2, qualifier=(1, "NL"), x12type=Money)
        tax_amount = ElementAccess(
            "AMT", 2, qualifier=(1, "T"), x12type=Money)
        total_claim_before_taxes = ElementAccess(
            "AMT", 2, qualifier=(1, "T2"), x12type=Money)

    class _ClaimAdjustments(X12LoopBridge):
        def __init__(self, aLoop, *args, **kwargs):
            super(Claim._ClaimAdjustments, self).__init__(
                aLoop, *args, **kwargs)
            self.contractual_obligation = ClaimAdjustment(
                aLoop, qualifier=(1, "CO"))
            self.correction_and_reversal = ClaimAdjustment(
                aLoop, qualifier=(1, "CR"))
            self.other = ClaimAdjustment(aLoop, qualifier=(1, "OA"))
            self.payor_initiated_reductions = ClaimAdjustment(
                aLoop, qualifier=(1, "PI"))
            self.patient_responsibility = ClaimAdjustment(
                aLoop, qualifier=(1, "PR"))

    loopName = "2100"

    # References
    group_or_policy_number = ElementAccess("REF", 2, qualifier=(1, "1L"))
    member_id = ElementAccess("REF", 2, qualifier=(1, "1W"))
    repriced_claim_reference_number = ElementAccess(
        "REF", 2, qualifier=(1, "9A"))
    adjusted_repriced_claim_reference_number = ElementAccess(
        "REF", 2, qualifier=(1, "9B"))
    employee_id = ElementAccess("REF", 2, qualifier=(1, "A6"))
    authorization_number = ElementAccess("REF", 2, qualifier=(1, "BB"))
    contract_class = ElementAccess("REF", 2, qualifier=(1, "CE"))
    medical_record_id = ElementAccess("REF", 2, qualifier=(1, "EA"))
    original_reference_number = ElementAccess("REF", 2, qualifier=(1, "F8"))
    prior_authorization_number = ElementAccess("REF", 2, qualifier=(1, "G1"))
    predetermination_of_benefits_number = ElementAccess(
        "REF", 2, qualifier=(1, "G3"))
    insurance_policy_number = ElementAccess("REF", 2, qualifier=(1, "IG"))
    ssn = ElementAccess("REF", 2, qualifier=(1, "SY"))

    provider_commercial_number = ElementAccess("REF", 2, qualifier=(1, "G2"))

    # Dates
    date_expiration = ElementAccess("DTM", 2, qualifier=(1, "036"), x12type=D8)
    date_received = ElementAccess("DTM", 2, qualifier=(1, "050"), x12type=D8)
    date_statement_period_start = ElementAccess(
        "DTM", 2, qualifier=(1, "232"), x12type=D8)
    date_statement_period_end = ElementAccess(
        "DTM", 2, qualifier=(1, "233"), x12type=D8)

    def __init__(self, anX12Message, *args, **kwargs):
        super(Claim, self).__init__(anX12Message, *args, **kwargs)
        self.claim_adjustments = self._ClaimAdjustments(anX12Message)
        self.line_items = self.loops(self._ServiceInfo, anX12Message)
        # Take advantage of ElementAccess attributes inheriting their parent's
        # qualifier. This needs to be fixed someday.
        self.patient = NamedEntity(anX12Message, qualifier=(1, "QC"))
        self.insured = NamedEntity(anX12Message, qualifier=(1, "IL"))
        self.corrected_insured = NamedEntity(anX12Message, qualifier=(1, "74"))
        self.service_provider = NamedEntity(anX12Message, qualifier=(1, "82"))
        self.crossover_carrier = NamedEntity(anX12Message, qualifier=(1, "TT"))
        self.corrected_priority_payer = NamedEntity(
            anX12Message, qualifier=(1, "PR"))
        self.payment_info = self._ClaimPaymentInfo(anX12Message)

    def get_actual_allowed_amount(self):
        s = Decimal('0.00')
        for li in self.line_items:
            s += li.get_actual_allowed_amount()
        return s

    def get_actual_deductible(self):
        s = self.claim_adjustments.patient_responsibility.total_amount('1')
        for li in self.line_items:
            s += li.get_actual_deductible()
        return s

    def get_actual_coinsurance(self):
        s = self.claim_adjustments.patient_responsibility.total_amount('2')
        for li in self.line_items:
            s += li.get_actual_coinsurance()
        return s

    def get_actual_copayment(self):
        s = self.claim_adjustments.patient_responsibility.total_amount('3')
        for li in self.line_items:
            s += li.get_actual_copayment()
        return s


class F835_4010(Facade):

    transaction_set_identifier_code = '835'
    x12_version_string = '4010'

    def __init__(self, anX12Message):
        """Examine the message and extract the relevant Loops."""
        st_loops = anX12Message.descendant('LOOP', name='ST_LOOP')
        if st_loops:
            self.facades = [F835_4010(loop) for loop in st_loops]
        else:
            self.header = first(self.loops(Header, anX12Message))
            self.payer = first(self.loops(Payer, anX12Message))
            self.payee = first(self.loops(Payee, anX12Message))
            self.claims_overview = first(
                self.loops(ClaimsOverview, anX12Message))
            self.claims = self.loops(Claim, anX12Message)
            self.footer = first(self.loops(Footer, anX12Message))


class F835_5010(Facade):

    transaction_set_identifier_code = '835'
    x12_version_string = '5010'

    def __init__(self, anX12Message):
        """Examine the message and extract the relevant Loops."""
        st_loops = anX12Message.descendant('LOOP', name='ST_LOOP')
        if st_loops:
            self.facades = [F835_5010(loop) for loop in st_loops]
        else:
            self.header = first(self.loops(Header, anX12Message))
            self.payer = first(self.loops(Payer, anX12Message))
            self.payee = first(self.loops(Payee, anX12Message))
            self.claims_overview = first(
                self.loops(ClaimsOverview, anX12Message))
            self.claims = self.loops(Claim, anX12Message)
            self.footer = first(self.loops(Footer, anX12Message))



#!/usr/bin/env python
"""The X12.map.source module emits Python source definitions from an :mod:`X12.parse` structure.

There are two flavors of Python source that can be produced:

    -   :class:`PythonVisitor` creates a single definition.
        A single Python constructor defines the entire message.

    -   :class:`FlatPythonVisitor` builds a flat definition.
        A number of individual constructors are assembled
        into a workable parser.

See :ref:`traversal` for notes on the **Visitor** design pattern.

..  autoclass:: PythonVisitor
    :members:

..  autoclass:: FlatPythonDetails
    :members:

..  autoclass:: FlatPythonVisitor
    :members:
"""

from tigershark.X12.parse import StructureVisitor
from tigershark.X12.parse import StopDescent

class PythonVisitor( StructureVisitor ):
    """Builds a single, huge definition.  Not very practical except
    for TINY messages.  Or if you set :py:data:`skipElement`
    to True, the resulting structure might be useful.

    :ivar skipElement: normally False, you can set this to True
        to produce a Segment-level message summary.
    """
    def __init__( self, varName, *args, **kw ):
        """Create a Python Visitor.
        :param varName: the name of the Python variable we'll product.
        """
        super( PythonVisitor, self ).__init__( *args, **kw )
        self.result= []
        self.varName= varName
        self.skipElement= False # change to True to prune tree, skipping elements
    def preMessage( self, msg, indent=0 ):
        self.result.append( "from tigershark.X12.parse import Message, Loop, Segment, Composite, Element, Properties" )
        if self.varName is None: self.varName= "x%s" % ( msg.name, )
        self.result.append( "%s = Message( %r, %r, " % ( self.varName, msg.name, msg.props ) )
    def postMessage( self, msg, indent=0 ):
        self.result.append( ")" )
    def preLoop( self, aLoop, indent=0 ):
        """Report this Loop.
        :param aLoop: a :class:`X12.parse.Loop` structure.
        :param indent: the indentation level.
        """
        self.result.append( indent*'  ' + "Loop( %r, %r, " % (aLoop.name, aLoop.props,) )
    def postLoop( self, aLoop, indent=0 ):
        self.result.append( indent*'  ' + ")," )
    def preSegment( self, aSegment, indent=0 ):
        """Report this Segment."""
        self.result.append( indent*'  ' + "Segment( %r, %r," % (aSegment.name, aSegment.props,) )
    def postSegment( self, aSegment, indent=0 ):
        self.result.append( indent*'  ' + ")," )
    def preComposite( self, aComposite, indent=0 ):
        """Report this Composite.  Configured by the :py:data:`skipElement` variable."""
        if self.skipElement:
            return
        self.result.append( indent*'  ' + "Composite( %r, %r," % (
            aComposite.name, aComposite.props ) )
    def postComposite( self, aComposite, indent=0 ):
        """Report this Composite.  Configured by the :py:data:`skipElement` variable."""
        if self.skipElement:
            return
        self.result.append( indent*'  ' + ")," )
    def postElement( self, anElement, indent=0 ):
        """Report this Element.  Configured by the :py:data:`skipElement` variable."""
        if self.skipElement:
            return
        elt= anElement
        self.result.append( indent*'  ' + "Element( %r, Properties(desc=%r, req_sit=%r, data_type=(%r,%r,%r), position=%d," % (
            elt.name, elt.desc, elt.req_sit, elt.type_name, elt.min_len, elt.max_len, elt.position ) )
        self.result.append( indent*'  ' + "  codes=%r ) )," % (
            elt.codes, ) )
    def getSource( self ):
        return "\n".join( self.result )

class FlatPythonDetails( PythonVisitor ):
    """Definitions, of Segment, Composite and Element.
    Messages and Loops are silently ignored.  They're handled by
    another visitor that scans the "top" of the structure.
    This Visitor is only used within a Loop or Message.
    """
    def preMessage( self, msg, indent= 0 ):
        pass
    def postMessage( self, msg, indent= 0 ):
        pass
    def preLoop( self, loop, indent=0 ):
        self.result.append( "%s_%s," % ( self.varName, loop.name, ) )
        raise StopDescent # prevent expansion of the Loop we just referenced
    def postLoop( self, loop, indent=0 ):
        pass

class FlatPythonVisitor( StructureVisitor ):
    """Decompose ALL Loops into separate definitions which can be referenced.
    This Visitor doesn't process Segments, Composites and Elements.
    It only looks at Messages and Loops.
    """
    def __init__( self, varName, *args, **kw ):
        super( FlatPythonVisitor, self ).__init__( *args, **kw )
        self.result= []
        self.varName= varName
    def preMessage( self, msg, indent=0 ):
        self.result.append( "from tigershark.X12.parse import Message, Loop, Segment, Composite, Element, Properties" )
    def postMessage( self, msg, indent=0 ):
        if self.varName is None: self.varName= "x%s" % ( msg.name, )
        self.result.append( "%s = Message( %r, %r, " % ( self.varName, msg.name, msg.props ) )
        for s in msg.structure:
            details= FlatPythonDetails( self.varName )
            # for Segment, Composites, etc., this involves a full pre-/post- visit
            # for Loops, however, it's a reference to a previous post- definition
            s.visit( details )
            self.result.extend( details.result )
        self.result.append( ")" )
    def preLoop( self, aLoop, indent=0 ):
        pass
    def postLoop( self, aLoop, indent=0 ):
        """Loop with References to stuff defined during pre-processing."""
        self.result.append( "%s_%s = Loop( %r, %r, " % ( self.varName, aLoop.name, aLoop.name, aLoop.props ) )
        for s in aLoop.structure:
            details= FlatPythonDetails( self.varName )
            s.visit( details )
            self.result.extend( details.result )
        self.result.append( ")" )
    def getSource( self ):
        return "\n".join( self.result )



"""Defines an insurance Claim from a business perspective,
mapping to underlying X12 structures.

There are two parts to this module.

-   A **Facade** over :mod:`X12.message` structures.
    See `Facade Use Cases`_ and `Facade Implementation`_.

-   A mapping of a Claim, using the **Facade** elements.
    See `Claim Mapping`_.


Claim Mapping
===============

**Patient** - Loop 2000C

..  autoclass:: Patient
..  autoclass:: Instutitional_Patient
..  autoclass:: Institutional_Inpatient
..  autoclass:: Institutional_Outpatient
..  autoclass:: Professional_Patient

**Provider** - Loop 2000A (a/k/a BillingProvider)

..  autoclass:: Provider

**Submitter** - Loop 1000A

..  autoclass:: Submitter

**Subscriber** - Loop 2000B

..  autoclass:: Subscriber

**Paperwork** -

..  autoclass:: Paperwork

**ClaimDetails** - Superclass for portions of Loop 2300, including embedded paperwork
  and various "HI" composite objects.

..  autoclass:: ClaimDetails

**Institutional_ClaimDetails**

..  autoclass:: Institutional_ClaimDetails

**Professional_ClaimDetails**

..  autoclass:: Professional_ClaimDetails

**ServiceLine**

..  autoclass:: ServiceLine
..  autoclass:: Institutional_ServiceLine
..  autoclass:: Professional_ServiceLine
..  autoclass:: ServiceLineAdj

Top-Level **Claim**.

..  autoclass:: Claim
"""
from __future__ import print_function

from tigershark.facade import X12LoopBridge
from tigershark.facade import X12SegmentBridge
from tigershark.facade import ElementAccess
from tigershark.facade import SegmentSequenceAccess
from tigershark.facade import SegmentConversion
from tigershark.facade import SequenceOf
from tigershark.facade import ElementSequenceAccess
from tigershark.facade import CompositeAccess
from tigershark.facade import CompositeSequenceAccess
from tigershark.facade import D8


class Patient( X12LoopBridge ):
    """Patient Data from loop 2000C"""
    loopName= "2000C"
    qualifier= ElementAccess( "NM1", 1 )
    entityType= ElementAccess( "NM1", 2 )
    last= ElementAccess( "NM1", 3 )
    first= ElementAccess( "NM1", 4 )
    mid= ElementAccess( "NM1", 5 )
    suffix= ElementAccess( "NM1", 7 )
    idQual= ElementAccess( "NM1", 8 )
    id= ElementAccess( "NM1", 9 )
    addr1= ElementAccess( "N3", 1 )
    addr2= ElementAccess( "N3", 2 )
    city= ElementAccess( "N4", 1 )
    state= ElementAccess( "N4", 2 )
    zip= ElementAccess( "N4", 3 )
    country= ElementAccess( "N4", 4 )
    dob= ElementAccess( "DMG", 2, x12type=D8 )
    sex= ElementAccess( "DMG", 3 )
    ssn= ElementAccess( "REF", 2, qualifier=(1,"SY") )
    def __str__( self ):
        return "%s, %s %s %s" % ( self.last, self.first, self.mid, self.suffix )

class Instutitional_Patient( Patient ):
    """Patient on an 837I claim."""
    rel= ElementAccess( "PAT", 1 )
    measCode= ElementAccess( "PAT", 7 )
    weight= ElementAccess( "PAT", 8 )
    resp= ElementAccess( "PAT", 9 )

class Institutional_Inpatient( Instutitional_Patient ):
    pass

class Institutional_Outpatient( Instutitional_Patient ):
    pass

class Professional_Patient( Patient ):
    """Patient on an 837P claim."""
    rel= ElementAccess( "PAT", 1 )
    dateFormat= ElementAccess( "PAT", 5 )
    dod= ElementAccess( "PAT", 6 )
    measCode= ElementAccess( "PAT", 7 )
    weight= ElementAccess( "PAT", 8 )
    preg= ElementAccess( "PAT", 9 )

class Provider( X12LoopBridge ):
    """Provider Data on the 2000A loop."""
    loopName= "2000A"
    qualifier= ElementAccess( "NM1", 1 )
    entityType= ElementAccess( "NM1", 2 )
    last= ElementAccess( "NM1", 3 )
    first= ElementAccess( "NM1", 4 )
    mid= ElementAccess( "NM1", 5 )
    nm106= ElementAccess( "NM1", 6 )
    suffix= ElementAccess( "NM1", 7 )
    idQual= ElementAccess( "NM1", 8 )
    taxId= ElementAccess( "NM1", 9 )
    npi= ElementAccess( "NM1", 9 )
    nm109= ElementAccess( "NM1", 9 )
    addr1= ElementAccess( "N3", 1 )
    addr2= ElementAccess( "N3", 2 )
    city= ElementAccess( "N4", 1 )
    state= ElementAccess( "N4", 2 )
    zip= ElementAccess( "N4", 3 )
    country= ElementAccess( "N4", 4 )
    contact_name= ElementAccess( "PER", 1 )
    contact_funct= ElementAccess( "PER", 2 )
    # Email, Phone and Fax are all on the same segment in pairs (3,4), (5,6), (7,8)
    # First value is a qualifier: "EM", "TE" or "FX"
    # Second value is the number
    email= ElementAccess( "PER", oneOf=( "EM", (3,4), (5,6), (7,8) ) )
    fax= ElementAccess( "PER", oneOf=( "FX", (3,4), (5,6), (7,8) ) )
    phone= ElementAccess( "PER", oneOf=( "TE", (3,4), (5,6), (7,8) ) )
    provCode= ElementAccess( "PRV", 1 )
    qual= ElementAccess( "PRV", 2 )
    taxonomyCd= ElementAccess( "PRV", 3 )
    BCID= ElementAccess( "REF", 2, qualifier=(1,"1A") )
    BSID= ElementAccess( "REF", 2, qualifier=(1,"1B") )
    medID= ElementAccess( "REF", 2, qualifier=(1,"1C") )
    uPIN= ElementAccess( "REF", 2, qualifier=(1,"1G") )
    commID= ElementAccess( "REF", 2, qualifier=(1,"G2") )
    taxID_sy= ElementAccess( "REF", 2, qualifier=(1,"SY") )
    taxID_ei= ElementAccess( "REF", 2, qualifier=(1,"EI") )
    ITSSpec= ElementAccess( "REF", 2, qualifier=(1,"N5") )
    legacyID= ElementAccess( "REF", 2, qualifier=(1,"X5") )
    def __str__( self ):
        return "%s, %s %s %s" % ( self.last, self.first, self.mid, self.suffix )

class Submitter( X12LoopBridge ):
    """Submitter Data derived from the 1000A loop."""
    loopName= "1000A"
    qualifier= ElementAccess( "NM1", 1 )
    entityType= ElementAccess( "NM1", 2 )
    last= ElementAccess( "NM1", 3 )
    first= ElementAccess( "NM1", 4 )
    mid= ElementAccess( "NM1", 5 )
    nm106= ElementAccess( "NM1", 6 )
    suffix= ElementAccess( "NM1", 7 )
    idQual= ElementAccess( "NM1", 8 )
    id= ElementAccess( "NM1", 9 )
    email= ElementAccess( "PER", oneOf=( "EM", (3,4), (5,6), (7,8) ) )
    fax= ElementAccess( "PER", oneOf=( "FX", (3,4), (5,6), (7,8) ) )
    phone= ElementAccess( "PER", oneOf=( "TE", (3,4), (5,6), (7,8) ) )
    def __str__( self ):
        return "%s, %s %s %s" % ( self.last, self.first, self.mid, self.suffix )

class Subscriber( X12LoopBridge ):
    """Subscriber Data from the 2000B loop."""
    loopName= "2000B"
    qualifier= ElementAccess( "NM1", 1 )
    entityType= ElementAccess( "NM1", 2 )
    last= ElementAccess( "NM1", 3 )
    first= ElementAccess( "NM1", 4 )
    mid= ElementAccess( "NM1", 5 )
    nm106= ElementAccess( "NM1", 6 )
    suffix= ElementAccess( "NM1", 7 )
    idQual= ElementAccess( "NM1", 8 )
    id= ElementAccess( "NM1", 9 )
    addr1= ElementAccess( "N3", 1 )
    addr2= ElementAccess( "N3", 2 )
    city= ElementAccess( "N4", 1 )
    state= ElementAccess( "N4", 2 )
    zip= ElementAccess( "N4", 3 )
    country= ElementAccess( "N4", 4 )
    payerResp= ElementAccess( "SBR", 1 )
    relCode= ElementAccess( "SBR", 2 )
    reference= ElementAccess( "SBR", 3 )
    name= ElementAccess( "SBR", 4 )
    insType= ElementAccess( "SBR", 5 )
    claimCode= ElementAccess( "SBR", 9 )
    rel= ElementAccess( "PAT", 1 )
    dateFormat= ElementAccess( "PAT", 5 )
    dod= ElementAccess( "PAT", 6 )
    measCode= ElementAccess( "PAT", 7 )
    weight= ElementAccess( "PAT", 8 )
    preg= ElementAccess( "PAT", 9 )
    dob= ElementAccess( "DMG", 2, x12type=D8 )
    gender= ElementAccess( "DMG", 3 )
    ssn= ElementAccess( "REF", 2, qualifier=(1,"SY") )
    def __str__( self ):
        return "%s, %s %s %s" % ( self.last, self.first, self.mid, self.suffix )

class Paperwork( X12SegmentBridge ):
    """Claim Supplemental Information, contained within ClaimDetails."""
    rptType= ElementAccess( "PWK", 1 )
    transCD= ElementAccess( "PWK", 2 )
    AttCtlQual= ElementAccess( "PWK", 6 )
    AttCtl= ElementAccess( "PWK", 7 )

class ClaimDetails( X12LoopBridge ):
    """Claim Details from the 2300 Loop."""
    loopName= "2300"
    patAcct= ElementAccess( "CLM", 1 )
    amount= ElementAccess( "CLM", 2 )
    serviceLoc= ElementAccess( "CLM", 5 )
    respCode1= ElementAccess( "CLM", 6 )
    medAssign= ElementAccess( "CLM", 7 )
    assignBenInd= ElementAccess( "CLM", 8 )
    release= ElementAccess( "CLM", 9 )
    signature= ElementAccess( "CLM", 10 )
    relCauses= ElementAccess( "CLM", 11 )
    progCode= ElementAccess( "CLM", 12 )
    provAgree= ElementAccess( "CLM", 16 )
    respCode2= ElementAccess( "CLM", 18 )
    delayCode= ElementAccess( "CLM", 20 )
    admsType= ElementAccess( "CL1", 1 )
    admsSrc= ElementAccess( "CL1", 2 )
    dischStat= ElementAccess( "CL1", 3 )
    coveredDays= ElementAccess( "QTY", 2, qualifier=(1,"CA") )
    coveredDays_UOM= ElementAccess( "QTY", 3, qualifier=(1,"CA") )
    nonCoveredDays= ElementAccess( "QTY", 2, qualifier=(1,"NA") )
    nonCoveredDays_UOM= ElementAccess( "QTY", 3, qualifier=(1,"NA") )
    coInsuranceDays= ElementAccess( "QTY", 2, qualifier=(1,"CD") )
    coInsuranceDays_UOM= ElementAccess( "QTY", 3, qualifier=(1,"CD") )
    reserveDays= ElementAccess( "QTY", 2, qualifier=(1,"LA") )
    reserveDays_UOM= ElementAccess( "QTY", 3, qualifier=(1,"LA") )
    origClaimNo= ElementAccess( "REF", 2, qualifier=(1,"F8") )
    medLRec= ElementAccess( "REF", 2, qualifier=(1,"EA") )
    microID= ElementAccess( "REF", 2, qualifier=(1,"D9") )
    paperwork= SegmentSequenceAccess( "PWK", x12type=SegmentConversion(Paperwork) )

class Institutional_ClaimDetails( ClaimDetails ):
    """Institutional Claim Details from the 2300 Loop."""
    fmt= ElementAccess( "DTP", 2 )
    stmtDts= ElementAccess( "DTP", 3, qualifier=(1,"434") )
    admsDt= ElementAccess( "DTP", 3, qualifier=(1,"435") )
    dischHr= ElementAccess( "DTP", 3, qualifier=(1,"096") )
    billNote= ElementAccess( "NTE", 2, qualifier=(1,"ADD") )
    # Non-"ADD" "NTE" segments are claim notes -- legacy uses an array of up to 10.
    # Only for institutional claims.
    claimNotes= ElementSequenceAccess( "NTE", 2, qualPos=1, notInList=("ADD",) )

    # DR Composite (Diagnosis Related Group (DRG) Information)
    drg= CompositeAccess( "HI", "DR", 1 )
    # BP/BR Composite (Principal Procedure Information)
    prinProcCode= CompositeAccess( "HI", ("BR","BP"), 1 )
    # BF Composite (Diagnosis) can have a number of attributes
    diagCode= CompositeSequenceAccess( "HI", "BF" )
    # BG Composite (Condition Information) can have a number of attributes
    condCode= CompositeSequenceAccess( "HI", "BG" )
    # BK Composite (Principal Diagnosis) is split into two attributes
    prinDiagCode= CompositeAccess( "HI", "BK", 1 )
    admitDiagCd= CompositeAccess( "HI", "BK", 2 )
    # BO/BQ Composite (Other Procedure Information) has a number of attributes
    procCode= CompositeSequenceAccess( "HI", ("BQ","BO") )
    # BH Composite (Occurrence Information) has a number of attributes
    occrCode= CompositeSequenceAccess( "HI", "BH" )
    # BE Composite (Value Information) has a number of attributes
    valueCode= CompositeSequenceAccess( "HI", "BE" )
    # BI Composite (Occurrence Span Information) is not used
    # TC Composite (Treatment Code Information) is not used

class Professional_ClaimDetails( ClaimDetails ):
    """Professional Claim Details from the 2300 Loop."""
    fmt= ElementAccess( "DTP", 2 )
    intDt= ElementAccess( "DTP", 3, qualifier=(1,"454") )
    accidentDt= ElementAccess( "DTP", 3, qualifier=(1,"439") )
    onsetDt= ElementAccess( "DTP", 3, qualifier=(1,"431") )
    lmpDt= ElementAccess( "DTP", 3, qualifier=(1,"484") )
    prescDt= ElementAccess( "DTP", 3, qualifier=(1,"471") )

    measCode= ElementAccess("CR1", 1)
    weight= ElementAccess("CR1", 2)
    ambTransCode= ElementAccess("CR1", 3)
    ambTransReason= ElementAccess("CR1", 4)
    ambTransQual= ElementAccess("CR1", 5)
    ambTransQty= ElementAccess("CR1", 6)
    ambDescript= ElementAccess("CR1", 9)
    ambDescript2= ElementAccess("CR1", 10)

    claimNotes= ElementSequenceAccess( "NTE", 2 )

    diagCode= ElementAccess( "HI", SequenceOf(1,8) )
    homeHealthCd= ElementAccess( "CR7", 1 )
    visitsRend= ElementAccess( "CR7", 2 )
    visitsProj= ElementAccess( "CR7", 3 )

    codeCat= ElementAccess("CRC", 1 )
    cond= ElementAccess("CRC", SequenceOf(2,6) )
    ref1Num= ElementAccess( "REF", 2, qualifier=(1,"9F") )
    preAuth= ElementAccess( "REF", 2, qualifier=(1,"G1") )

class ServiceLine( X12LoopBridge ):
    """Service Line from the 2400 Loop."""
    loopName= "2400"
    SrvLNo= ElementAccess( "LX", 1 )
    format= ElementAccess( "DTP", 2 )
    srvDT= ElementAccess( "DTP", 3, qualifier=(1,"472") )
    dmeCode= ElementAccess( "SV5", 1 )
    qualifier= ElementAccess( "SV5", 2 )
    amt1= ElementAccess( "SV5", 3 )
    amt2= ElementAccess( "SV5", 4 )
    amt3= ElementAccess( "SV5", 5 )
    amt4= ElementAccess( "SV5", 6 )

class Institutional_ServiceLine( ServiceLine ):
    """Institutional Service Line from the 2400 Loop."""
    revenueCode= ElementAccess( "SV2", 1 )
    procID= ElementAccess( "SV2", 2 )
    amt= ElementAccess( "SV2", 3 )
    unitsQual= ElementAccess( "SV2", 4 )
    units= ElementAccess( "SV2", 5 )
    unitRate= ElementAccess( "SV2", 6 )
    amtNotCovered= ElementAccess( "SV2", 7 )

class Professional_ServiceLine( ServiceLine ):
    """Professional Service Line from the 2400 Loop."""
    srvID= ElementAccess( "SV1", 1 )
    amt= ElementAccess( "SV1", 2 )
    unitsQual= ElementAccess( "SV1", 3 )
    units= ElementAccess( "SV1", 4 )
    facCode= ElementAccess( "SV1", 5 )
    diagCode= ElementAccess( "SV1", 7 )
    respCode1= ElementAccess( "SV1", 9 )
    respCode2= ElementAccess( "SV1", 11 )
    respCode3= ElementAccess( "SV1", 12 )
    coPay= ElementAccess( "SV1", 15 )

class ServiceLineAdj( X12LoopBridge ):
    """Service Line Adjustment from the 2430 Loop."""
    loopName= "2430"
    idCode= ElementAccess( "SVD", 1 )
    amt= ElementAccess( "SVD", 2 )
    procID= ElementAccess( "SVD", 3 )
    svcID= ElementAccess( "SVD", 4 )
    qty= ElementAccess( "SVD", 5 )
    assNo= ElementAccess( "SVD", 6 )
    format= ElementAccess( "DTP", 2 )
    srvDT= ElementAccess( "DTP", 3, qualifier=(1,"579") )

class Claim( object ):
    """A claim, built from an X12 :samp:`837` message.

    :ivar submitter: A sequence of :class:`Submitter` instances from the 1000A loop.
    :ivar provider: A sequence of :class:`Provider` instances from the 2000A loop.
    :ivar subscriber: A sequence of :class:`Subscriber` instances from the 2000B loop.
    :ivar patient: A sequence of :class:`Institutional_Inpatient` instances from the 2000C loop.
    """
    def loops( self, theClass, anX12Message ):
        return [ theClass(loop) for loop in anX12Message.descendant( "loop", theClass.loopName ) ]

    def __init__( self, anX12Message ):
        """Examine the message and extract the relevant Loops."""
        # XXX - determine 837P vs. 837I and Inpatient vs. Outpatient
        self.submitter= self.loops( Submitter, anX12Message )
        print("Loop 1000A, Submitter", [str(x) for x in self.submitter ])
        self.provider= self.loops(Provider, anX12Message)
        print("Loop 2000A, Billing Provider", [str(x) for x in self.provider])
        self.subscriber= self.loops(Subscriber, anX12Message)
        print("Loop 2000B, Subscriber", [str(x) for x in self.subscriber])
        self.patient= self.loops( Institutional_Inpatient, anX12Message )
        print("Loop 2000C, Patient", [str(x) for x in self.patient])
        # Loop 2300 - Claim Details
        self.claimDetails= self.loops( Institutional_ClaimDetails, anX12Message )
        print("Loop 2300, Claim Details", [str(x) for x in self.claimDetails])
        # Loop 2400 - Service Lines



"""Define :mod:`X12.parse` structures for some standard X12 Segments.

These are definitions of some standard Segments without a lot
of supporting detail.  This can be used (or extended) to cover
cases where X12 segment metadata is not available.

This module includes a conversion to emit "better-than-nothing"
:class:`X12.parse.Segment` definitions for a few standard
X12 Segments.

:var isa: ISA Segment
:var gs: GS Segment
:var st: ST Segment
:var se: SE Segment
:var ge: GE Segment
:var iea: IEA Segment

..  autofunction:: buildSegment
"""
from __future__ import print_function

from tigershark.X12.parse import Element
from tigershark.X12.parse import Properties
from tigershark.X12.parse import Segment

ISAFields = """\
ISA01 Authorization Information Qualifier
ISA02 Authorization Information
ISA03 Security Information Qualifier
ISA04 Security Information
ISA05 Interchange ID Qualifier
ISA06 Interchange Sender ID
ISA07 Interchange ID Qualifier
ISA08 Interchange Receiver ID
ISA09 Interchange Date
ISA10 Interchange Time
ISA11 Interchange Control Standards ID
ISA12 Interchange Control Version Number
ISA13 Interchange Control Number
ISA14 Acknowledgement Requested
ISA15 Test Indicator
ISA16 Subelement Separator
"""

GSFields = """\
GS01  Functional ID code
GS02  Application Sender's Code
GS03  Application Receiver's Code
GS04  Date
GS05  Time
GS06  Group Control Number
GS07  Responsible Agency Code
GS08  Version/Rel. Ind. ID Code"""

STFields = """\
ST01  Transaction set ID code
ST02  Transaction set control number
"""

# SE02 must match ST02
SEFields = """\
SE01  Number of included segments
SE02  Transaction set control number"""

# GE02 must match GS06
GEFields = """\
GE01  Number of Transaction Sets Included in this Function Group
GE02  Group Control Number"""

# IEA02 must match ISA13
IEAFields = """\
IEA01 Number of Included Functional Groups
IEA02 Interchange Control Number (same as ISA13)
"""


def buildSegment(name, desc, fieldText):
    """Convert a block of text to a Segment definition.
    The block of text has one Element defined on each line.
    The Element definition is an Element name, one or more spaces, and the
    element description.

    :param name: the name of the segment
    :param desc: the description of the segment
    :param fieldText: a block of text, with one Element name
        and description per line.
    :returns: :class:`X12.parse.Segment` definition
    """
    segment = Segment(name, Properties(desc=desc))
    for line in filter(None, fieldText.splitlines()):
        name, description = line.strip().split(None, 1)
        element = Element(name, Properties(desc=description, codes=[]))
        segment.append(element)
    return segment

isa = buildSegment("ISA", "ISA", ISAFields)
gs = buildSegment("GS", "GS", GSFields)
st = buildSegment("ST", "ST", STFields)
se = buildSegment("SE", "SE", SEFields)
ge = buildSegment("GE", "GE", GEFields)
iea = buildSegment("IEA", "IEA", IEAFields)
